﻿/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "analyze.h"


/**
 * @description: 将ast::Col转换为TabCol
 * @param {const std::shared_ptr<ast::Col>&} col
 * @return {TabCol}
 */
TabCol convertColName(std::shared_ptr<ast::Col>& col,std::map<std::string,int> &aliasMap){
  TabCol tabCol;
  tabCol.tab_name = col->tab_name;
  tabCol.col_name = col->col_name;
  if(!col->alias.empty()){
    tabCol.alias = col->alias;
  }else if(col->isAgg){
 switch (col->aggType) {
   case ast::AggFuncType::SUM:
     tabCol.alias = "SUM(" + tabCol.col_name + ")";
     break;
   case ast::AggFuncType::COUNT:
     tabCol.alias = "COUNT(" + tabCol.col_name + ")";
     break;
   case ast::AggFuncType::COUNT_ALL:
     tabCol.alias = "COUNT(*)";
     break;
   case ast::AggFuncType::AVG:
     tabCol.alias = "AVG(" + tabCol.col_name + ")";
     break;
   case ast::AggFuncType::MIN:
     tabCol.alias = "MIN(" + tabCol.col_name + ")";
     break;
   case ast::AggFuncType::MAX:
     tabCol.alias = "MAX(" + tabCol.col_name + ")";
     break;
   }
  }else {
    tabCol.alias = tabCol.col_name;
  }
  aliasMap[tabCol.alias]++;
  if(aliasMap[tabCol.alias] > 1){
    tabCol.alias = tabCol.alias + std::to_string(aliasMap[tabCol.alias]);
  }
  return tabCol;
}



/**
 * @description: 分析器，进行语义分析和查询重写，需要检查不符合语义规定的部分
 * @param {shared_ptr<ast::TreeNode>} parse parser生成的结果集
 * @return {shared_ptr<Query>} Query
 */
std::shared_ptr<Query> Analyze::do_analyze(std::shared_ptr<ast::TreeNode> parse)
{
    std::shared_ptr<Query> query = std::make_shared<Query>();
    if (auto x = std::dynamic_pointer_cast<ast::SelectStmt>(parse))
    {
        // 处理表名
        query->tables = std::move(x->tabs);
        /** TODO: 检查表是否存在 */
        for(auto &tab_name : query->tables){
            if(sm_manager_->fhs_.find(tab_name) == sm_manager_->fhs_.end()){//如果表不存在
                throw TableNotFoundError(tab_name);
            }
        }

        query->hasAgg = x->has_agg();
        query->hasGroup = x->has_group_by;


        // 处理target list，再target list中添加上表名，例如 a.id
        //加入到query cols中
        std::map<std::string,int> aliasMap;
        for (int i=0;i<x->cols.size();i++) {
             auto sel_col = convertColName(x->cols[i],aliasMap);
            //如果有聚合函数或者group by
            if(query->hasAgg || x->group_by){
              //如果不是聚合函数
              if(!x->cols[i]->isAgg){
                if(!query->hasGroup) {
                  throw InternalError("Invalid Group Function");
                }
                //检查该列是不是group by的列
                bool flag =false;
                for(auto &gCol : x->group_by->cols){
                  if(gCol->col_name == x->cols[i]->col_name){
                    std::map<std::string,int> noneMap;
                    TabCol t  = convertColName(x->cols[i],noneMap);
                    query->groupByCols.push_back(t);
                    flag = true;
                    break;
                  }
                }
                if(!flag){
                  throw ColumnNotGroupByError(x->cols[i]->col_name);
                }
              }else{//如果是聚合函数
                AggCol aggCol;
                //按顺序添加到聚合列中
                if(sel_col.col_name == "*") {
                  //如果不是count类型
                  if(x->cols[i]->aggType != ast::AggFuncType::COUNT){
                    throw InternalError("Invalid agg function");
                  }
                  aggCol.AggType = ast::AggFuncType::COUNT_ALL;
                  sel_col.tab_name = "*";
                  sel_col.col_name = "*";
                }
                aggCol.col = sel_col;
                aggCol.AggType = x->cols[i]->aggType;
                query->aggCols.push_back(aggCol);
              }
            }
            //Normal
            query->cols.push_back(sel_col);
        }


        std::vector<ColMeta> all_cols;
        get_all_cols(query->tables, all_cols);
        if (query->cols.empty()) {
            // select all columns SELECT * 的情况
            for (auto &col : all_cols) {
              TabCol sel_col = {.tab_name = col.tab_name, .col_name = col.name};
              query->cols.push_back(sel_col);
              //检查group by是否齐全
              if(query->hasGroup){
                for(auto &col : all_cols){
                  bool flag = false;
                  for(auto &gCol : x->group_by->cols){
                    if(col.name == gCol->col_name){
                      flag = true;
                      query->groupByCols.push_back(sel_col);
                      break;
                    }
                  }
                  if(!flag){
                    throw ColumnNotGroupByError(col.name);
                  }
                }
              }
            }
        } else {
            // infer table name from column name
            for (auto &sel_col : query->cols) {
                //放行count*
                if(sel_col.col_name == "*"){
                  continue;
                }
                sel_col = check_column(all_cols, sel_col);  // 列元数据校验
            }
        }
        //处理where条件
        get_clause(x->conds, query->conds);
        check_clause(query->tables, query->conds);
        //处理having条件
        if(x->has_group_by && !x->group_by->conds.empty()){
            get_having_clause(x->group_by->conds, query->havingConds);
            check_having_clause(query->tables, query->havingConds);
        }
    } else if (auto x = std::dynamic_pointer_cast<ast::UpdateStmt>(parse)) {
        /** TODO: */
        //处理where条件
        get_clause(x->conds, query->conds);
        check_clause({x->tab_name}, query->conds);
        //处理set值
        for (auto &sv_set : x->set_clauses) {
            SetClause set_clause;
            set_clause.lhs = {.tab_name = x->tab_name, .col_name = sv_set->col_name};
            set_clause.rhs = convert_sv_value(sv_set->val);
            query->set_clauses.push_back(set_clause);
        }
    } else if (auto x = std::dynamic_pointer_cast<ast::DeleteStmt>(parse)) {
        //处理where条件
        get_clause(x->conds, query->conds);
        check_clause({x->tab_name}, query->conds);
    } else if (auto x = std::dynamic_pointer_cast<ast::InsertStmt>(parse)) {
        // 处理insert 的values值
        for (auto &sv_val : x->vals) {
            query->values.push_back(convert_sv_value(sv_val));
        }
    } else {
        // do nothing
    }
    query->parse = std::move(parse);
    return query;
}


TabCol Analyze::check_column(const std::vector<ColMeta> &all_cols, TabCol target) {
    //检查字段是否能在表中找到
    if (target.tab_name.empty()) {
        // Table name not specified, infer table name from column name
        std::string tab_name;
        for (auto &col : all_cols) {
            if (col.name == target.col_name) {
                if (!tab_name.empty()) {
                    throw AmbiguousColumnError(target.col_name);
                }
                tab_name = col.tab_name;
            }
        }
        if (tab_name.empty()) {
            throw ColumnNotFoundError(target.col_name);
        }
        target.tab_name = tab_name;
    } else {
        /** TODO: Make sure target column exists */
        bool found = false;
        for (auto &col : all_cols) {
            if (col.tab_name == target.tab_name && col.name == target.col_name) {
                found = true;
                break;
            }
        }
        if (!found) {
            throw ColumnNotFoundError(target.col_name);
        }
    }
     return  target;
}

void Analyze::get_all_cols(const std::vector<std::string> &tab_names, std::vector<ColMeta> &all_cols) {
    for (auto &sel_tab_name : tab_names) {
        // 这里db_不能写成get_db(), 注意要传指针
        const auto &sel_tab_cols = sm_manager_->db_.get_table(sel_tab_name).cols;
        all_cols.insert(all_cols.end(), sel_tab_cols.begin(), sel_tab_cols.end());
    }
}


void Analyze::get_clause(const std::vector<std::shared_ptr<ast::BinaryExpr>> &sv_conds, std::vector<Condition> &conds) {
    conds.clear();
    for (auto &expr : sv_conds) {
        if(expr->lhs->isAgg) {//如果where中有agg字段
          throw AggInWhereError();
        }
        if(expr->rhs != nullptr && expr->rhs->isAgg) {//如果where中有agg字段
          throw AggInWhereError();
        }
        Condition cond;
        //如果有子查询
        if(expr->subquery != nullptr) {
          cond.subquery_query = do_analyze(expr->subquery);
        }
        cond.isIn = expr->isIn;
        //ast in Value list 转 Common Value,Raw后面检查完毕后再设置
        for(auto &val : expr->rlist){
          cond.in_values.push_back(std::make_shared<Value>(convert_sv_value(val)));
        }
        //如果有in，两种情况，一种是常量列表，一种是子查询
        //如果is_rhs_val为true的话后面就不会检查列表名
        cond.lhs_col = {.tab_name = expr->lhs->tab_name, .col_name = expr->lhs->col_name};
        cond.op = convert_sv_comp_op(expr->op);
        if(cond.subquery_query != nullptr || !expr->rlist.empty()){
            cond.is_rhs_val = true;
        }else if (auto rhs_val = std::dynamic_pointer_cast<ast::Value>(expr->rhs)){
            cond.is_rhs_val = true;
            cond.rhs_val = convert_sv_value(rhs_val);
        } else if (auto rhs_col = std::dynamic_pointer_cast<ast::Col>(expr->rhs)) {
            cond.is_rhs_val = false;
            cond.rhs_col = {.tab_name = rhs_col->tab_name, .col_name = rhs_col->col_name};
        }
      conds.push_back(cond);
    }
}

void Analyze::get_having_clause(const std::vector<std::shared_ptr<ast::BinaryExpr>> &sv_conds, std::vector<HavingConditon> &conds) {
    conds.clear();
    for (auto &expr : sv_conds) {
        HavingConditon cond;
        //如果having中有agg字段
        if(expr->lhs->isAgg) {
          cond.is_agg = true;
          cond.ag_type = expr->lhs->aggType;
        }
        cond.lhs_col = {.tab_name = expr->lhs->tab_name, .col_name = expr->lhs->col_name};
        cond.op = convert_sv_comp_op(expr->op);
        if (auto rhs_val = std::dynamic_pointer_cast<ast::Value>(expr->rhs)) {
            cond.is_rhs_val = true;
            cond.rhs_val = convert_sv_value(rhs_val);
        } else if (auto rhs_col = std::dynamic_pointer_cast<ast::Col>(expr->rhs)) {
            cond.is_rhs_val = false;
            cond.rhs_col = {.tab_name = rhs_col->tab_name, .col_name = rhs_col->col_name};
        }
      conds.push_back(cond);
    }
}

void Analyze::check_clause(const std::vector<std::string> &tab_names, std::vector<Condition> &conds) {
    // auto all_cols = get_all_cols(tab_names);
    std::vector<ColMeta> all_cols;
    get_all_cols(tab_names, all_cols);
    // Get raw values in where clause
    for (auto &cond : conds) {
        // Infer table name from column name
        cond.lhs_col = check_column(all_cols, cond.lhs_col);
        TabMeta &lhs_tab = sm_manager_->db_.get_table(cond.lhs_col.tab_name);
        auto lhs_col = lhs_tab.get_col(cond.lhs_col.col_name);
        ColType lhs_type = lhs_col->type;
        ColType rhs_type;
        //如果cond右值是常量
        if (cond.is_rhs_val && cond.in_values.empty()) {
            cond.rhs_val.init_raw(lhs_col->len);
            rhs_type = cond.rhs_val.type;
        } else if(cond.subquery_query == nullptr && cond.in_values.empty()){//否则就来表里找有没有这个字段,前提不是子查询
            if (!cond.is_rhs_val ) {
                cond.rhs_col = check_column(all_cols, cond.rhs_col);
            }
            TabMeta &rhs_tab = sm_manager_->db_.get_table(cond.rhs_col.tab_name);
            auto rhs_col = rhs_tab.get_col(cond.rhs_col.col_name);
            rhs_type = rhs_col->type;
        }else {
            //只用判断in常量列表即可，后面子查询查询出类型不对会报错
            for(auto &val : cond.in_values){
              if(lhs_type == TYPE_INT && val->type == TYPE_FLOAT){
                // int和float可以比较
                val->init_raw(lhs_col->len);
              }else if(lhs_type == TYPE_FLOAT && val->type == TYPE_INT) {
                val->init_raw(lhs_col->len);
                // int和float可以比较
              }else if(lhs_type == val->type){//
                val->init_raw(lhs_col->len);
              }else {
                throw IncompatibleTypeError(coltype2str(lhs_type), coltype2str(val->type));
              }
            }
            return;
        }
        if (lhs_type != rhs_type) {
          if(lhs_type == TYPE_INT && rhs_type == TYPE_FLOAT){
            // int和float可以比较
          }else if(lhs_type == TYPE_FLOAT && rhs_type == TYPE_INT) {
            // int和float可以比较
          }else if(cond.subquery_query == nullptr && cond.in_values.empty()){//没有子查询才扔错误
            throw IncompatibleTypeError(coltype2str(lhs_type), coltype2str(rhs_type));
          }
        }
    }
}

void Analyze::check_having_clause(const std::vector<std::string> &tab_names, std::vector<HavingConditon> &conds) {
  // auto all_cols = get_all_cols(tab_names);
  std::vector<ColMeta> all_cols;
  get_all_cols(tab_names, all_cols);
  // Get raw values in where clause
  for (auto &cond : conds) {
    if(cond.lhs_col.col_name == "*" && cond.ag_type == ast::AggFuncType::COUNT) {
      continue;
    }
    // Infer table name from column name
    cond.lhs_col = check_column(all_cols, cond.lhs_col);
    if (!cond.is_rhs_val) {
      cond.rhs_col = check_column(all_cols, cond.rhs_col);
    }
    TabMeta &lhs_tab = sm_manager_->db_.get_table(cond.lhs_col.tab_name);
    auto lhs_col = lhs_tab.get_col(cond.lhs_col.col_name);
    ColType lhs_type = lhs_col->type;
    if(cond.ag_type == ast::AggFuncType::COUNT){//如果是count就是整数
      lhs_type = TYPE_INT;
    }
    ColType rhs_type;
    if (cond.is_rhs_val) {
      cond.rhs_val.init_raw(lhs_col->len);
      rhs_type = cond.rhs_val.type;
    } else {
      TabMeta &rhs_tab = sm_manager_->db_.get_table(cond.rhs_col.tab_name);
      auto rhs_col = rhs_tab.get_col(cond.rhs_col.col_name);
      rhs_type = rhs_col->type;
    }
    if (lhs_type != rhs_type) {
      if(lhs_type == TYPE_INT && rhs_type == TYPE_FLOAT){
        // int和float可以比较
      }else if(lhs_type == TYPE_FLOAT && rhs_type == TYPE_INT) {
        // int和float可以比较
      }else{
        throw IncompatibleTypeError(coltype2str(lhs_type), coltype2str(rhs_type));
      }
    }
  }
}


Value Analyze::convert_sv_value(const std::shared_ptr<ast::Value> &sv_val) {
    Value val;
    if (auto int_lit = std::dynamic_pointer_cast<ast::IntLit>(sv_val)) {
        val.set_int(int_lit->val);
    } else if (auto float_lit = std::dynamic_pointer_cast<ast::FloatLit>(sv_val)) {
        val.set_float(float_lit->val);
    } else if (auto str_lit = std::dynamic_pointer_cast<ast::StringLit>(sv_val)) {
        val.set_str(str_lit->val);
    } else {
        throw InternalError("Unexpected sv value type");
    }
    return val;
}

CompOp Analyze::convert_sv_comp_op(ast::SvCompOp op) {
    std::map<ast::SvCompOp, CompOp> m = {
        {ast::SV_OP_EQ, OP_EQ}, {ast::SV_OP_NE, OP_NE}, {ast::SV_OP_LT, OP_LT},
        {ast::SV_OP_GT, OP_GT}, {ast::SV_OP_LE, OP_LE}, {ast::SV_OP_GE, OP_GE},
    };
    return m.at(op);
}
