/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once

#include <cassert>
#include <cstring>
#include <ctime>
#include <iomanip>
#include <iostream>
#include <memory>
#include <sstream>
#include <string>
#include "defs.h"
#include "parser/ast.h"
#include "record/rm_defs.h"
#include "system/sm_meta.h"


// 定义颜色
#define RESET   "\033[0m"
#define RED     "\033[31m"      // 红色
#define GREEN   "\033[32m"      // 绿色
#define YELLOW  "\033[33m"       // 黄色
#define BLUE "\033[34m"         // 蓝色


class Query;class AbstractExecutor;class Plan;class Analyze;inline
std::string current_time() {
  auto now = std::chrono::system_clock::now();
  auto in_time_t = std::chrono::system_clock::to_time_t(now);
  std::ostringstream ss;
  ss << std::put_time(std::localtime(&in_time_t), "%Y-%m-%d %X");
  return ss.str();
}

// 定义日志级别
#define LOG_DEBUG(message) \
do { \
std::ostringstream oss; \
oss << BLUE << "[DEBUG] [" << current_time() << "] " << message << RESET; \
std::cout << oss.str() << std::endl; \
} while (0)

#define LOG_INFO(message) \
do { \
std::ostringstream oss; \
oss << GREEN << "[INFO] [" << current_time() << "] " << message << RESET; \
std::cout << oss.str() << std::endl; \
} while (0)

#define LOG_WARN(message) \
do { \
std::ostringstream oss; \
oss << YELLOW << "[WARN] [" << current_time() << "] " << message << RESET; \
std::cout << oss.str() << std::endl; \
} while (0)

#define LOG_ERROR(message) \
do { \
std::ostringstream oss; \
oss << RED << "[ERROR] [" << current_time() << "] " << message << RESET; \
std::cout << oss.str() << std::endl; \
} while (0)


struct TabCol {
  std::string tab_name;
  std::string col_name;
  std::string alias;

  friend bool operator<(const TabCol& x, const TabCol& y) {
    return std::make_pair(x.tab_name, x.col_name) < std::make_pair(y.tab_name, y.col_name);
  }
};

struct Value {
  ColType type; // type of value
  union {
    int int_val; // int value
    float float_val; // float value
  };

  std::string str_val; // string value

  std::shared_ptr<RmRecord> raw; // raw record buffer

    Value(const ColType& t,const char* val,int len){
        type = t;
        switch (type) {
          case TYPE_INT:
            assert(len == sizeof(int));
            int_val = *(int *)(val);
            break;
          case TYPE_FLOAT:
            assert(len == sizeof(float));
            float_val = *(float *)(val);
            break;
          case TYPE_STRING:
            str_val = std::string(val,len);
            break;
          default:
            // 可以在这里处理其他类型或抛出异常
            break;
        }
    }

    explicit Value(){
      type = TYPE_NONE;
      int_val = 0;
      float_val = 0;
      str_val = "";
    }

    void set_int(int int_val_) {
        type = TYPE_INT;
        int_val = int_val_;
    }

    void set_float(float float_val_) {
        type = TYPE_FLOAT;
        float_val = float_val_;
    }

    void set_str(std::string str_val_) {
        type = TYPE_STRING;
        str_val = std::move(str_val_);
    }

    size_t get_len() const{
      if (type == TYPE_INT) {
        return sizeof(int);
      } else if (type == TYPE_FLOAT) {
        return sizeof(float);
      } else if (type == TYPE_STRING) {
        return str_val.size();
      } else {
        // throw std::runtime_error("Unknown type");
        return 0;
      }
    }

    char* get_data(){
      if (type == TYPE_INT) {
        return (char*)&int_val;
      } else if (type == TYPE_FLOAT) {
        return (char*)&float_val;
      } else if (type == TYPE_STRING) {
        return (char*)str_val.c_str();
      } else {
        // throw std::runtime_error("Unknown type");
        return nullptr;
      }
    }



    void init_raw(int len) {
//        assert(raw == nullptr);
        raw = std::make_shared<RmRecord>(len);
        if (type == TYPE_INT) {
            assert(len == sizeof(int));
            *(int *)(raw->data) = int_val;
        } else if (type == TYPE_FLOAT) {
            assert(len == sizeof(float));
            *(float *)(raw->data) = float_val;
        } else if (type == TYPE_STRING) {
            if (len < (int)str_val.size()) {
                throw StringOverflowError();
            }
            memset(raw->data, 0, len);
            memcpy(raw->data, str_val.c_str(), str_val.size());
        }
    }

    std::size_t hash() const {
      std::size_t seed = 0;
      // 根据 Value 的类型和值计算哈希值
      switch (type) {
        case TYPE_INT:
          seed ^= std::hash<int>{}(int_val) + 0x9e3779b9 + (seed << 6) + (seed >> 2);
          break;
        case TYPE_FLOAT:
          seed ^= std::hash<float>{}(float_val) + 0x9e3779b9 + (seed << 6) + (seed >> 2);
          break;
        case TYPE_STRING:
          seed ^= std::hash<std::string>{}(str_val) + 0x9e3779b9 + (seed << 6) + (seed >> 2);
          break;
        case TYPE_NONE:
          throw InternalError("Unknown type");
      }
      return seed;
    }

    bool operator==(const Value& other) const {
      if (type != other.type) {
        return false;
      }
      switch (type) {
        case TYPE_INT:
          return int_val == other.int_val;
        case TYPE_FLOAT:
          return float_val == other.float_val;
        case TYPE_STRING:
          return str_val == other.str_val;
          // 其他类型...
        case TYPE_NONE:
          throw InternalError("Unknown type");
      }
      return false;
    }




};

struct AggValue : public Value {
  //辅助计算AVG
  bool minFlag = false;
  bool maxFlag = false;
  std::string colName;
  AggValue(const ColType& t,const char* val,int len):Value(t,val,len){
  }
  AggValue(){
  }
};

enum CompOp { OP_EQ, OP_NE, OP_LT, OP_GT, OP_LE, OP_GE };

struct Condition {
    TabCol lhs_col;   // left-hand side column
    CompOp op;        // comparison operator
    bool is_rhs_val = false;  // true if right-hand side is a value (not a column)
    TabCol rhs_col;   // right-hand side column
    Value rhs_val;    // right-hand side value
    bool is_agg = false;
    //Subquery
    bool isIn = false;
    std::shared_ptr<Query> subquery_query;
    std::shared_ptr<Plan> subquery_plan;
    std::shared_ptr<AbstractExecutor> subquery_exec;
    std::vector<std::shared_ptr<Value>> in_values;

};

struct SetClause {
    TabCol lhs;
    Value rhs;
};

struct AggCol{
  ast::AggFuncType AggType;
  TabCol col;
  std::string alias;
};

struct AggColMeta{
  ast::AggFuncType ag_type;
  ColMeta col;
  std::string alias;
};

struct  HavingConditon: Condition {
  ast::AggFuncType ag_type;
  AggColMeta agg_col;
};
