/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class SortExecutor : public AbstractExecutor {
private:
  std::unique_ptr<AbstractExecutor> prev_;
  ColMeta cols_; // 框架中只支持一个键排序，需要自行修改数据结构支持多个键排序
  size_t tuple_num;
  bool is_desc_;
  std::vector<size_t> used_tuple;
  std::vector<std::unique_ptr<RmRecord>> tuples_;
  int position;
  std::unique_ptr<RmRecord> current_tuple;

public:
  SortExecutor(std::unique_ptr<AbstractExecutor> prev, TabCol sel_cols, bool is_desc) {
    prev_ = std::move(prev);
    cols_ = prev_->get_col_offset(sel_cols);
    is_desc_ = is_desc;
    tuple_num = 0;
    used_tuple.clear();
    tuples_.clear();
    position = 0;
  }

  void beginTuple() override {
    used_tuple.clear();
    for (prev_->beginTuple(); !prev_->is_end(); prev_->nextTuple()) {
      auto cur_tuple = prev_->Next();
      tuples_.emplace_back(std::move(cur_tuple));
      tuple_num++;
    }
    std::sort(tuples_.begin(), tuples_.end(),
              [this](const std::unique_ptr<RmRecord>& a, const std::unique_ptr<RmRecord>& b) {
                auto a_data = a->data + cols_.offset;
                auto b_data = b->data + cols_.offset;
                if (cols_.type == TYPE_INT) {
                  int a_val = *reinterpret_cast<int*>(a_data);
                  int b_val = *reinterpret_cast<int*>(b_data);
                  return is_desc_ ? a_val > b_val : a_val < b_val;
                }
                if (cols_.type == TYPE_FLOAT) {
                  float a_val = *reinterpret_cast<float*>(a_data);
                  float b_val = *reinterpret_cast<float*>(b_data);
                  return is_desc_ ? a_val > b_val : a_val < b_val;
                }
                char* a_val = a_data;
                char* b_val = b_data;
                return is_desc_ ? strcmp(a_val, b_val) > 0 : strcmp(a_val, b_val) < 0;
              });
    position = 0;
  }

  void nextTuple() override {
    position++;
  }

  std::unique_ptr<RmRecord> Next() override {
    return std::move(tuples_[position]);
  }

  const std::vector<ColMeta>& cols()  const override {
    return prev_->cols();
  }

  bool is_end() const override {
    return position >= static_cast<int>(tuples_.size());
  }

  Rid& rid() override { return _abstract_rid; }
};
