/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once

#include "execution_defs.h"
#include "common/common.h"
#include "index/ix.h"
#include "system/sm.h"

class AbstractExecutor {
   public:
    Rid _abstract_rid;

    Context *context_;

    virtual ~AbstractExecutor() = default;

    virtual size_t tupleLen() const { return 0; };

    virtual const std::vector<ColMeta> &cols() const {
        std::vector<ColMeta> *_cols = nullptr;
        return *_cols;
    };

    virtual std::string getType() { return "AbstractExecutor"; };

    virtual void beginTuple(){};

    virtual void nextTuple(){};

    virtual bool is_end() const { return true; };

    virtual Rid &rid() = 0;

    virtual std::unique_ptr<RmRecord> Next() = 0;

    virtual ColMeta get_col_offset(const TabCol &target) { return ColMeta();};

    std::vector<ColMeta>::const_iterator get_col(const std::vector<ColMeta> &rec_cols, const TabCol &target) {
        auto pos = std::find_if(rec_cols.begin(), rec_cols.end(), [&](const ColMeta &col) {
            return col.tab_name == target.tab_name && col.name == target.col_name ;
        });
        if (pos == rec_cols.end()) {
            throw ColumnNotFoundError(target.tab_name + '.' + target.col_name);
        }
        return pos;
    }

    std::vector<ColMeta>::const_iterator get_agg_col(const std::vector<ColMeta> &rec_cols, const TabCol &target) {
      auto pos = std::find_if(rec_cols.begin(), rec_cols.end(), [&](const ColMeta &col) {
        return col.tab_name == target.tab_name && col.name == target.col_name && col.ailas == target.alias;
      });
      if (pos == rec_cols.end()) {
        throw ColumnNotFoundError(target.tab_name + '.' + target.col_name);
      }
      return pos;
    }

    bool CompareCol(const Condition &cond, const ColMeta &leftCol, ColType left_t, char *left,
                    ColType right_t, char *right) {
      int cmp = 0;
      bool res = false;
      if (left_t == ColType::TYPE_INT && right_t == ColType::TYPE_FLOAT) {
        // 将左转化成float
        char *left_float = new char[4];
        *(float *)left_float = *(int *)left;
        cmp = ix_compare(left_float, right, right_t, leftCol.len);
        delete[] left_float;
      } else if (left_t == ColType::TYPE_FLOAT && right_t == ColType::TYPE_INT) {
        // 将右转化成float
        char *right_float = new char[4];
        *(float *)right_float = *(int *)right;
        cmp = ix_compare(left, right_float, left_t, leftCol.len);
        delete[] right_float;
      } else if (left_t != right_t) {
        throw InternalError("Type mismatch");
      } else {
        cmp = ix_compare(left, right, right_t, leftCol.len);
      }
      // ix比较器
      if (cond.op == OP_EQ) {
        res = cmp == 0;
      } else if (cond.op == OP_NE) {
        res = cmp != 0;
      } else if (cond.op == OP_LT) {
        res = cmp < 0;
      } else if (cond.op == OP_GT) {
        res = cmp > 0;
      } else if (cond.op == OP_LE) {
        res = cmp <= 0;
      } else if (cond.op == OP_GE) {
        res = cmp >= 0;
      } else {
        throw InternalError("Unexpected op type");
      }
      return res;
    }
  bool Check_Conditions(const std::vector<ColMeta>& recCols,const std::vector<Condition> &conds, const std::unique_ptr<RmRecord> &rec) {
      //对每个cond验证
      bool res = true;
      for (auto &cond : conds) {
          auto condCol = cond.lhs_col;
          //找到条件字段在记录字段中的位置
          auto &leftCol = *get_col(recCols, condCol);

          ColType left_t = leftCol.type;
          char *left = rec->data + leftCol.offset;

          ColType right_t;
          char *right;

          //Subquery Record
          std::unique_ptr<RmRecord> subquery_rec;

          if (cond.is_rhs_val && cond.subquery_query == nullptr && !cond.isIn) {
              //如果是常量
              right_t = cond.rhs_val.type;
              right = cond.rhs_val.raw->data;
          }else if(cond.subquery_exec != nullptr && !cond.isIn) {//子查询
              cond.subquery_exec->beginTuple();
              subquery_rec= cond.subquery_exec->Next();
              cond.subquery_exec->nextTuple();
              auto cols = cond.subquery_exec->cols();
              if(subquery_rec == nullptr) return false;
              if(!cond.subquery_exec->is_end()) {
                  throw InternalError("Subquery return more than one row");
              }
              if(cols.size()>1) {
                  throw InternalError("Subquery return more than one column");
              }
              right_t = cols[0].type;
              right = subquery_rec->data;
          }else if(cond.subquery_exec != nullptr && cond.isIn) {
              //IN子查询
              bool in_res = false;
              cond.subquery_exec->beginTuple();
              while(!cond.subquery_exec->is_end()) {
                  subquery_rec = cond.subquery_exec->Next();
                  if(subquery_rec == nullptr) return false;
                  auto cols = cond.subquery_exec->cols();
                  if(cols.size()>1) {
                      throw InternalError("Subquery return more than one column");
                  }
                 if(CompareCol(cond,leftCol,left_t,left,cols[0].type,subquery_rec->data)) {
                    in_res = true;
                    break;
                 }
                  cond.subquery_exec->nextTuple();
              }
              if(!in_res) return false;
              continue;
          }else if(cond.isIn && !cond.in_values.empty()) {
              bool in_res = false;
              for(auto& val : cond.in_values) {
                  if(CompareCol(cond,leftCol,left_t,left,val->type,val->raw->data)) {
                      in_res = true;
                      break;
                  }
              }
              if(!in_res) return false;
              continue;
          }
          else{
            //如果是字段
            auto rightCol = *get_col(recCols, cond.rhs_col);
          right = rec->data + rightCol.offset;
          right_t = rightCol.type;
          }
          res = CompareCol(cond,leftCol,left_t,left,right_t,right);
          if(!res) break;
      }
      return res;
    }

};
