/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once

#include <cfloat>
#include <climits>

#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class IndexScanExecutor : public AbstractExecutor {
private:
  std::string tab_name_; // 表名称
  TabMeta tab_; // 表的元数据
  std::vector<Condition> conds_; // 扫描条件
  RmFileHandle* fh_; // 表的数据文件句柄
  std::vector<ColMeta> cols_; // 需要读取的字段
  size_t len_; // 选取出来的一条记录的长度
  std::vector<Condition> fed_conds_; // 扫描条件，和conds_字段相同
  bool is_end_ = false;

  std::vector<std::string> index_col_names_; // index scan涉及到的索引包含的字段
  IndexMeta index_meta_; // index scan涉及到的索引元数据

  Rid rid_;
  std::unique_ptr<RecScan> scan_;

  SmManager* sm_manager_;

public:
  IndexScanExecutor(SmManager* sm_manager, std::string tab_name, std::vector<Condition> conds,
                    std::vector<std::string> index_col_names, Context* context) {
    sm_manager_ = sm_manager;
    context_ = context;
    tab_name_ = std::move(tab_name);
    tab_ = sm_manager_->db_.get_table(tab_name_);
    conds_ = std::move(conds);
    // index_no_ = index_no;
    index_col_names_ = index_col_names;
    index_meta_ = *(tab_.get_index_meta(index_col_names_));
    fh_ = sm_manager_->fhs_.at(tab_name_).get();
    cols_ = tab_.cols;
    len_ = cols_.back().offset + cols_.back().len;
    std::map<CompOp, CompOp> swap_op = {
      {OP_EQ, OP_EQ}, {OP_NE, OP_NE}, {OP_LT, OP_GT}, {OP_GT, OP_LT}, {OP_LE, OP_GE}, {OP_GE, OP_LE},
    };

    for (auto& cond : conds_) {
      if (cond.lhs_col.tab_name != tab_name_) {
        // lhs is on other table, now rhs must be on this table
        assert(!cond.is_rhs_val && cond.rhs_col.tab_name == tab_name_);
        // swap lhs and rhs
        std::swap(cond.lhs_col, cond.rhs_col);
        cond.op = swap_op.at(cond.op);
      }
    }
    fed_conds_ = conds_;
  }

  void beginTuple() override {
    #ifdef DEBUG
    LOG_DEBUG("IndexScanExecutor beginTuple");
    #endif
    context_->lock_mgr_->lock_shared_on_table(context_->txn_, fh_->GetFd());
    // 初始化索引扫描
    auto idx_name = sm_manager_->get_ix_manager()->get_index_name(tab_name_, index_col_names_);
    auto ih = sm_manager_->ihs_.at(idx_name).get();
    is_end_ = ih->is_empty();
    if (is_end_) {
      return;
    }
    auto [max_rc, min_rc] = get_max_and_min_rc();

    // 初始化index scan
    auto lower_bound = ih->lower_bound(min_rc);
    auto upper_bound = ih->upper_bound(max_rc);
    delete[] max_rc;
    delete[] min_rc;

    #ifdef DEBUG
    LOG_DEBUG("lower_bound: " + std::to_string(lower_bound.page_no) + "  " + std::to_string(lower_bound.slot_no));
    LOG_DEBUG("upper_bound: " + std::to_string(upper_bound.page_no) + "  " + std::to_string(upper_bound.slot_no));
    #endif

    // 开始scan
    auto bpm = sm_manager_->get_bpm();
    for (scan_ = std::make_unique<IxScan>(ih, lower_bound, upper_bound, bpm); !is_end(); scan_->next()) {
      auto rec = fh_->get_record(scan_->rid(), context_);

      if (Check_Conditions(cols_, fed_conds_, rec)) {
        break;
      }
    }
  }

  void nextTuple() override {
    for (scan_->next(); !is_end(); scan_->next()) {
      auto rec = fh_->get_record(scan_->rid(), context_);
      if (Check_Conditions(cols_, fed_conds_, rec)) {
        break;
      }
    }
  }

  std::unique_ptr<RmRecord> Next() override {
    rid_ = scan_->rid();
    return fh_->get_record(rid_, context_);
  }

  bool is_end() const override {
    return is_end_ || scan_->is_end();
  }

  Rid& rid() override {
    rid_ = scan_->rid();
    return rid_;
  }

  size_t tupleLen() const override {
    return len_;
  }

  ColMeta get_col_offset(const TabCol& target) override {
    auto pos = get_col(cols_, target);
    return *pos;
  }

  const std::vector<ColMeta>& cols() const override {
    return cols_;
  };

  /*
  * @brief 判断索引的第一个字段是否为用户期望的字段，用于归并连接中判断是否有序
  */
  bool user_expected_index(std::string idx_name) {
    auto first_idx_name = index_col_names_[0];
    return idx_name == first_idx_name;
  }

private:
  std::pair<char*, char*> get_max_and_min_rc() {
    // 初始化max_rc和min_rc
    char* max_rc = new char[index_meta_.col_tot_len];
    char* min_rc = new char[index_meta_.col_tot_len];
    init_max_and_min_rc(max_rc, min_rc);

    // 通过条件更新max_rc和min_rc
    int off_set = 0;
    auto idx_name = sm_manager_->get_ix_manager()->get_index_name(tab_name_, index_col_names_);
    for (int i = 0; i < index_meta_.col_num; i++) {
      auto col_name = index_meta_.cols[i].name;
      auto col_len = index_meta_.cols[i].len;
      auto col_type = index_meta_.cols[i].type;
      for (const auto& cond : conds_) {
        if (cond.lhs_col.col_name == col_name && cond.is_rhs_val) {
          handle_condition(cond, max_rc, min_rc, off_set, col_len, col_type);
        }
      }
      off_set += index_meta_.cols[i].len;
    }

    return {max_rc, min_rc};
  }

  void handle_condition(Condition cond, char* max_rc, char* min_rc, int off_set, int len, ColType type) {
    auto op = cond.op;
    auto rhs_data = cond.rhs_val.raw->data;
    if (op == OP_EQ) {
      if (ix_compare(rhs_data, max_rc + off_set, type, len) < 0) {
        memcpy(max_rc + off_set, rhs_data, len);
      }
      if (ix_compare(rhs_data, min_rc + off_set, type, len) > 0) {
        memcpy(min_rc + off_set, rhs_data, len);
      }
    } else if (op == OP_GT || op == OP_GE) {
      if (ix_compare(rhs_data, min_rc + off_set, type, len) > 0) {
        memcpy(min_rc + off_set, rhs_data, len);
      }
    } else if (op == OP_LT || op == OP_LE) {
      if (ix_compare(rhs_data, max_rc + off_set, type, len) < 0) {
        memcpy(max_rc + off_set, rhs_data, len);
      }
    } else if (op == OP_NE) {
    } else {
      throw UnknowenTypeError();
    }
  }

  void init_max_and_min_rc(char* max_rc, char* min_rc) {
    int off_set = 0;
    for (size_t i = 0; i < index_meta_.col_num; i++) {
      set_type_max(max_rc + off_set, index_meta_.cols[i]);
      set_type_min(min_rc + off_set, index_meta_.cols[i]);
      off_set += index_meta_.cols[i].len;
    }
  }

  void set_type_max(char* data, const ColMeta& col) {
    switch (col.type) {
    case TYPE_INT:
      *(int*)data = INT_MAX;
      break;
    case TYPE_FLOAT:
      *(float*)data = FLT_MAX;
      break;
    case TYPE_STRING:
      memset(data, 0xff, col.len);
      break;
    default:
      throw UnknowenTypeError();
    }
  }

  void set_type_min(char* data, const ColMeta& col) {
    switch (col.type) {
    case TYPE_INT:
      *(int*)data = INT_MIN;
      break;
    case TYPE_FLOAT:
      *(float*)data = FLT_MIN;
      break;
    case TYPE_STRING:
      memset(data, 0, col.len);
      break;
    default:
      throw UnknowenTypeError();
    }
  }
};
