/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class InsertExecutor : public AbstractExecutor {
private:
  TabMeta tab_; // 表的元数据
  std::vector<std::vector<Value>> values_; // 需要插入的数据
  RmFileHandle* fh_; // 表的数据文件句柄
  std::string tab_name_; // 表名称
  Rid rid_; // 插入的位置，由于系统默认插入时不指定位置，因此当前rid_在插入后才赋值
  SmManager* sm_manager_;

public:
  InsertExecutor(SmManager* sm_manager, const std::string& tab_name, std::vector<std::vector<Value>> values, Context* context) {
    sm_manager_ = sm_manager;
    tab_ = sm_manager_->db_.get_table(tab_name);
    values_ = values;
    tab_name_ = tab_name;
    if (values[0].size() != tab_.cols.size()) {
      throw InvalidValueCountError();
    }
    fh_ = sm_manager_->fhs_.at(tab_name).get();
    context_ = context;
  };

  std::unique_ptr<RmRecord> Next() override {
    //对表上写锁
    if (!context_->lock_mgr_->lock_exclusive_on_table(context_->txn_, fh_->GetFd())) {
      throw TransactionAbortException(context_->txn_->get_transaction_id(), AbortReason::UPGRADE_CONFLICT);
    }
    for(auto &values_vec : values_) {

    // Make record buffer
      RmRecord rec(fh_->get_file_hdr().record_size);
      for (size_t i = 0; i < values_vec.size(); i++) {
        auto& col = tab_.cols[i];
        auto& val = values_vec[i];
        if (col.type != val.type) {
          if (col.type == TYPE_FLOAT && val.type == TYPE_INT) {
            val.set_float(static_cast<float>(val.int_val));
          } else {
            throw IncompatibleTypeError(coltype2str(col.type), coltype2str(val.type));
          }
        }
        val.init_raw(col.len);
        memcpy(rec.data + col.offset, val.raw->data, col.len);
      }

      //检查是否重复
      std::vector<Rid> res;
      for (auto& index_meta : tab_.indexes) {
        auto ih = sm_manager_->ihs_.at(sm_manager_->get_ix_manager()->get_index_name(tab_name_, index_meta.cols)).get();

        char* key = new char[index_meta.col_tot_len];
        int offset = 0;
        for (size_t i = 0; i < index_meta.col_num; ++i) {
          memcpy(key + offset, rec.data + index_meta.cols[i].offset, index_meta.cols[i].len);
          offset += index_meta.cols[i].len;
        }

        if (ih->get_value(key, &res, context_->txn_)) {
          delete[] key;
          throw InternalError("Duplicate key for unique index !");
        }

        delete[] key;
      }
      // Insert into record file
      rid_ = fh_->insert_record(rec.data, context_);

      //记录到事物的insert日志
      auto* write_record = new WriteRecord(WType::INSERT_TUPLE, tab_name_, rid_);
      context_->txn_->append_record_write_record(write_record);


      //添加到redo log
      // redoLogAppend(rec);

      // Insert into index
      for (size_t i = 0; i < tab_.indexes.size(); ++i) {
        auto& index = tab_.indexes[i];
        auto ih = sm_manager_->ihs_.at(sm_manager_->get_ix_manager()->get_index_name(tab_name_, index.cols)).get();
        char* key = new char[index.col_tot_len];
        int offset = 0;
        for (size_t i = 0; i < index.col_num; ++i) {
          memcpy(key + offset, rec.data + index.cols[i].offset, index.cols[i].len);
          offset += index.cols[i].len;
        }
        ih->insert_entry(key, rid_, context_->txn_);

        auto index_write_record = new IndexWriteRecord(WType::INSERT_TUPLE, tab_name_, rid_, index.col_tot_len, key);
        context_->txn_->append_index_write_record(index_write_record);

        delete[] key;
      }
    }

    return nullptr;
  }

  // void redoLogAppend(RmRecord rec)  {
  //   // InsertLogRecord record(context_->txn_->get_transaction_id(), context_->txn_->get_prev_lsn(), rec, rid_, tab_name_);
  //   // lsn_t lsn = context_->log_mgr_->add_log_to_buffer(&record);
  //   context_->txn_->set_prev_lsn(lsn);
  //   Page* page = sm_manager_->get_bpm()->fetch_page({fh_->GetFd(), rid_.page_no});
  //   page->set_page_lsn(lsn);
  //   sm_manager_->get_bpm()->unpin_page({fh_->GetFd(), rid_.page_no}, true);
  // }

  Rid& rid() override { return rid_; }
};
