/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include <queue>

#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class SortMergeJoinExecutor : public AbstractExecutor {
  friend class MergeSortComparator;

private:
  DiskManager *disk_manager_;
  std::unique_ptr<AbstractExecutor> left_; // 左儿子节点（需要join的表）
  std::unique_ptr<AbstractExecutor> right_; // 右儿子节点（需要join的表）
  size_t len_; // join后获得的每条记录的长度
  std::vector<ColMeta> cols_; // join后获得的记录的字段

  std::vector<Condition> fed_conds_; // join条件
  bool is_end_;
  std::unique_ptr<RmRecord> left_record_;
  std::unique_ptr<RmRecord> right_record_;

  std::vector<std::unique_ptr<RmRecord>> left_cartesian_buffer_;
  std::vector<std::unique_ptr<RmRecord>> right_cartesian_buffer_;
  int cur_left_cartesian_idx_ = -1;
  int cur_right_cartesian_idx_ = -1;

  size_t sort_buffer_max_size_;
  std::vector<std::unique_ptr<RmRecord>> sort_buffer_;
  int left_buffer_file_cnt_ = 0;
  int right_buffer_file_cnt_ = 0;
  std::string uuid_;
  std::string left_sorted_filename_;
  std::string right_sorted_filename_;
  std::ifstream left_ifs;
  std::ifstream right_ifs;

public:
  SortMergeJoinExecutor(DiskManager *disk_manager, std::unique_ptr<AbstractExecutor> left,
                        std::unique_ptr<AbstractExecutor> right,
                        std::vector<Condition> conds) {
    disk_manager_ = disk_manager;
    left_ = std::move(left);
    right_ = std::move(right);
    len_ = left_->tupleLen() + right_->tupleLen();
    cols_ = left_->cols();
    auto right_cols = right_->cols();
    for (auto &col : right_cols) {
      col.offset += static_cast<int>(left_->tupleLen());
    }

    cols_.insert(cols_.end(), right_cols.begin(), right_cols.end());
    is_end_ = false;
    fed_conds_ = std::move(conds);

    sort_buffer_max_size_ = 3000;
    uuid_ = get_milltimestamp_str();
#ifdef DEBUG
    LOG_DEBUG("SortMergeJoinExecutor::uuid_ :" + uuid_);
#endif

    left_sorted_filename_ = "left-" + uuid_ + "-sorted";
    right_sorted_filename_ = "right-" + uuid_ + "-sorted";
  }

  bool is_end() const override {
    return is_end_;
  };

  size_t tupleLen() const override {
    return len_;
  };

  const std::vector<ColMeta> &cols() const override {
    return this->cols_;
  };


  void beginTuple() override {
#ifdef DEBUG
    LOG_DEBUG("SortMergeJoinExecutor::beginTuple()");
#endif
    is_end_ = false;

    //如果需要排序的话进行排序
    make_subexecutor_sorted(left_.get(), true);
    make_subexecutor_sorted(right_.get(), false);

    //打印"sorted_results"文件
    print_sorted_results();

    //现在已经得到了两个有序的二进制文件，下面进行归并排序
    left_ifs.open(left_sorted_filename_, std::ios::binary);
    right_ifs.open(right_sorted_filename_, std::ios::binary);

    if (!left_ifs.is_open() || !right_ifs.is_open()) {
      throw InternalError("Failed to open sorted file when executing SortMergeJoinExecutor::beginTuple()");
    }
    left_record_ = get_record_from_ifs(&left_ifs, left_.get());
    right_record_ = get_record_from_ifs(&right_ifs, right_.get());
    nextTuple();
  }

  void nextTuple() override {
    if (left_record_ == nullptr || right_record_ == nullptr) {
      if (cur_left_cartesian_idx_ >= 0 && cur_right_cartesian_idx_ >= 0) {
        if (cur_right_cartesian_idx_ < right_cartesian_buffer_.size() - 1) {
          cur_right_cartesian_idx_++;
          return;
        }
        if (cur_left_cartesian_idx_ < left_cartesian_buffer_.size() - 1) {
          cur_left_cartesian_idx_++;
          cur_right_cartesian_idx_ = 0;
          return;
        }

        cur_left_cartesian_idx_ = -1;
        cur_right_cartesian_idx_ = -1;
        left_cartesian_buffer_.clear();
        right_cartesian_buffer_.clear();
      }
      is_end_ = true;
      return;
    }
    if (is_end_) {
      return;
    }
    if (cur_left_cartesian_idx_ >= 0 && cur_right_cartesian_idx_ >= 0) {
      if (cur_right_cartesian_idx_ < right_cartesian_buffer_.size() - 1) {
        cur_right_cartesian_idx_++;
        return;
      }
      if (cur_left_cartesian_idx_ < left_cartesian_buffer_.size() - 1) {
        cur_left_cartesian_idx_++;
        cur_right_cartesian_idx_ = 0;
        return;
      }
      cur_left_cartesian_idx_ = -1;
      cur_right_cartesian_idx_ = -1;
      left_cartesian_buffer_.clear();
      right_cartesian_buffer_.clear();
    }
    while (left_record_ != nullptr && right_record_ != nullptr) {
      int cmp = compare(left_record_, right_record_);
      if (cmp == 0) {
        //处理笛卡尔积
        auto pre_record = std::make_unique<RmRecord>(*right_record_);

        while (left_record_ != nullptr && compare(left_record_, pre_record) == 0) {
          left_cartesian_buffer_.emplace_back(std::make_unique<RmRecord>(*left_record_));
          left_record_ = get_record_from_ifs(&left_ifs, left_.get());
        }

        while (right_record_ != nullptr && compare(right_record_, pre_record) == 0) {
          right_cartesian_buffer_.emplace_back(std::make_unique<RmRecord>(*right_record_));
          right_record_ = get_record_from_ifs(&right_ifs, right_.get());
        }

        cur_left_cartesian_idx_ = 0;
        cur_right_cartesian_idx_ = 0;
        return;
      } else if (cmp < 0) {
        left_record_ = get_record_from_ifs(&left_ifs, left_.get());
      } else {
        right_record_ = get_record_from_ifs(&right_ifs, right_.get());
      }
    }
  }

  std::unique_ptr<RmRecord> Next() override {
    if (is_end_) {
      throw InternalError("SortMergeJoinExecutor::Next() called when is_end_ is true");
    }

    auto res = std::make_unique<RmRecord>(len_);

    if (cur_left_cartesian_idx_ >= 0 && cur_right_cartesian_idx_ >= 0) {
      auto tmp_l_rec = left_cartesian_buffer_[cur_left_cartesian_idx_].get();
      auto tmp_r_rec = right_cartesian_buffer_[cur_right_cartesian_idx_].get();
      memcpy(res->data, tmp_l_rec->data, tmp_l_rec->size);
      memcpy(res->data + tmp_l_rec->size, tmp_r_rec->data, tmp_r_rec->size);
    } else {
      memcpy(res->data, left_record_->data, left_record_->size);
      memcpy(res->data + left_record_->size, right_record_->data, right_record_->size);
    }

    return res;
  }

  ColMeta get_col_offset(const TabCol &target) override {
    auto pos = get_col(cols_, target);
    return *pos;
  }

  Rid &rid() override { return _abstract_rid; }

private:
  //调用的时候确保ifs已经打开，而且不是eof
  std::unique_ptr<RmRecord> get_record_from_ifs(std::ifstream *ifs, AbstractExecutor *executor) {
    if (ifs->eof()) {
      return nullptr;
    }
    auto record = std::make_unique<RmRecord>(executor->tupleLen());
    for (const auto &col : executor->cols()) {
      ifs->read(record->data + col.offset, col.len);
    }
    if (ifs->eof()) {
      return nullptr;
    }
    return std::move(record);
  }

  void make_subexecutor_sorted(AbstractExecutor *executor, bool is_left) {
    auto filename_prefix = (is_left ? "left-" : "right-") + uuid_ + "-";
    auto col_name = is_left ? fed_conds_[0].lhs_col.col_name : fed_conds_[0].rhs_col.col_name;
    if (need_sort(executor, col_name)) {
#ifdef DEBUG
      LOG_DEBUG("need sort when executing SortMergeJoinExecutor::make_subexecutor_sorted()");
#endif

      auto cnt = 0;
      //将下一层执行器的所有数据读入缓冲区
      for (executor->beginTuple(); !executor->is_end(); executor->nextTuple()) {
        //如果缓冲区满了就需要排序后存入文件
        if (cnt >= sort_buffer_max_size_) {
          std::string filename = filename_prefix + std::to_string(
                                     is_left ? left_buffer_file_cnt_ : right_buffer_file_cnt_);
          std::sort(sort_buffer_.begin(), sort_buffer_.end(),
                    [=](const std::unique_ptr<RmRecord> &left, const std::unique_ptr<RmRecord> &right) {
                      return compare_for_sort_buffer(left, right, executor, is_left) < 0;
                    });
          write_sorted_clips_to_file(executor, filename, sort_buffer_);
          is_left ? left_buffer_file_cnt_++ : right_buffer_file_cnt_++;
          sort_buffer_.clear();
          cnt = 0;
        }
        auto record = executor->Next();
        sort_buffer_.push_back(std::move(record));
        cnt++;
      }
      std::string filename = filename_prefix + std::to_string(
                                 is_left ? left_buffer_file_cnt_ : right_buffer_file_cnt_);
      std::sort(sort_buffer_.begin(), sort_buffer_.end(),
                [=](const std::unique_ptr<RmRecord> &left, const std::unique_ptr<RmRecord> &right) {
                  return compare_for_sort_buffer(left, right, executor, is_left) < 0;
                });
      write_sorted_clips_to_file(executor, filename, sort_buffer_);
      is_left ? left_buffer_file_cnt_++ : right_buffer_file_cnt_++;
      sort_buffer_.clear();

      //得到多个文件，对他们进行归并排序，放入大文件中
      auto out_filename = is_left ? left_sorted_filename_ : right_sorted_filename_;
      merge_sorted_files(filename_prefix, out_filename, executor, is_left);

      //这行代码没有用，只是为了让不使用索引更慢一点
      // std::this_thread::sleep_for(std::chrono::seconds(2));

      return;
    }
    //如果不需要排序，直接将数据写入文件
    auto out_filename = is_left ? left_sorted_filename_ : right_sorted_filename_;
    std::ofstream ofs(out_filename, std::ios::binary);

    for (executor->beginTuple(); !executor->is_end(); executor->nextTuple()) {
      auto record = executor->Next();

      //写入out文件
      for (auto &col : executor->cols()) {
        ofs.write(record->data + col.offset, col.len);
      }
    }
    ofs.close();
  }

  std::string get_milltimestamp_str() {
    auto now = std::chrono::system_clock::now();
    auto now_ms = std::chrono::time_point_cast<std::chrono::milliseconds>(now);
    return std::to_string(now_ms.time_since_epoch().count());
  }

  int compare_for_sort_buffer(const std::unique_ptr<RmRecord> &left, const std::unique_ptr<RmRecord> &right,
                              AbstractExecutor *executor, bool is_left) {
    auto col_meta = *get_col(executor->cols(), is_left ? fed_conds_[0].lhs_col : fed_conds_[0].rhs_col);
    auto type = col_meta.type;
    auto left_data = left->data + col_meta.offset;
    auto right_data = right->data + col_meta.offset;
    switch (type) {
      case TYPE_INT: {
        int left_val = *reinterpret_cast<int *>(left_data);
        int right_val = *reinterpret_cast<int *>(right_data);
        if (left_val == right_val)
          return 0;
        return left_val - right_val > 0 ? 1 : -1;
      }
      case TYPE_FLOAT: {
        float left_val = *reinterpret_cast<float *>(left_data);
        float right_val = *reinterpret_cast<float *>(right_data);
        if (left_val == right_val)
          return 0;
        return left_val - right_val > 0 ? 1 : -1;
      }
      case TYPE_STRING: {
        return strcmp(left_data, right_data);
      }
      default: {
        throw InternalError("Unsupported column type when executing SortMergeJoinExecutor::compare()");
      }
    }
  }

  void merge_sorted_files(const std::string &filename_prefix, const std::string &out_filename,
                          AbstractExecutor *executor, bool is_left) {
    const int file_cnt = is_left ? left_buffer_file_cnt_ : right_buffer_file_cnt_;
    auto comparator = [=](const std::pair<std::unique_ptr<RmRecord>, size_t> &left,
                          const std::pair<std::unique_ptr<RmRecord>, size_t> &right) -> bool {
      return compare_for_sort_buffer(left.first, right.first, executor, is_left) > 0;
    };
    std::priority_queue<
      std::pair<std::unique_ptr<RmRecord>, size_t>,
      std::vector<std::pair<std::unique_ptr<RmRecord>, size_t>>,
      decltype(comparator)> min_heap(comparator);
    std::vector<std::ifstream> inter_files(file_cnt);

    //打开所有中间文件，读取第一条记录放入堆中
    for (int i = 0; i < file_cnt; i++) {
      //打开中间文件，如果打开失败或打开即EOF则抛出异常
      inter_files[i].open(filename_prefix + std::to_string(i), std::ios::binary);
      if (!inter_files[i].is_open() || inter_files[i].eof()) {
        throw InternalError(
            "Failed to open intermediate file or eof when executing SortMergeJoinExecutor::merge_sorted_files()");
      }
      //读取第一条记录放入堆中
      auto record = std::make_unique<RmRecord>(executor->tupleLen());
      for (const auto &col : executor->cols()) {
        inter_files[i].read(record->data + col.offset, col.len);
      }
      min_heap.emplace(std::move(record), i);
    }

    //文件之前不会存在，不需要用app打开
    std::ofstream ofs(out_filename, std::ios::binary);
    while (!min_heap.empty()) {
      auto record = std::move(const_cast<std::unique_ptr<RmRecord> &>(min_heap.top().first));
      auto file_idx = min_heap.top().second;
      min_heap.pop();

      //写入out文件
      for (auto &col : executor->cols()) {
        ofs.write(record->data + col.offset, col.len);
      }

      if (!inter_files[file_idx].eof()) {
        auto rec_tmp = std::make_unique<RmRecord>(executor->tupleLen());
        for (const auto &col : executor->cols()) {
          inter_files[file_idx].read(rec_tmp->data + col.offset, col.len);
        }
        if (!inter_files[file_idx].eof()) {
          min_heap.emplace(std::move(rec_tmp), file_idx);
        }
      }
    }

    for (auto &file : inter_files) {
      file.close();
    }
    ofs.close();
  }

  int compare(const std::unique_ptr<RmRecord> &left, const std::unique_ptr<RmRecord> &right) {
    auto left_col_meta = *get_col(left_->cols(), fed_conds_[0].lhs_col);
    auto right_col_meta = *get_col(right_->cols(), fed_conds_[0].rhs_col);
    if (left_col_meta.type != right_col_meta.type) {
      throw InternalError("Join column type mismatch when executing SortMergeJoinExecutor::compare()");
    }
    auto type = left_col_meta.type;
    auto left_data = left->data + left_col_meta.offset;
    auto right_data = right->data + right_col_meta.offset;

    switch (type) {
      case TYPE_INT: {
        int left_val = *reinterpret_cast<int *>(left_data);
        int right_val = *reinterpret_cast<int *>(right_data);
        if (left_val == right_val)
          return 0;
        return left_val - right_val > 0 ? 1 : -1;
      }
      case TYPE_FLOAT: {
        float left_val = *reinterpret_cast<float *>(left_data);
        float right_val = *reinterpret_cast<float *>(right_data);
        if (left_val == right_val)
          return 0;
        return left_val - right_val > 0 ? 1 : -1;
      }
      case TYPE_STRING: {
        return strcmp(left_data, right_data);
      }
      default: {
        throw InternalError("Unsupported column type when executing SortMergeJoinExecutor::compare()");
      }
    }
  }

  int create_intermediate_result_file_for_merge_sort(const std::string &filename) {
    disk_manager_->create_file(filename);
    auto fd = disk_manager_->open_file(filename);
    return fd;
  }

  void write_sorted_clips_to_file(AbstractExecutor *executor, const std::string &filename,
                                  std::vector<std::unique_ptr<RmRecord>> &sort_buffer_) {
    std::ofstream ofs(filename, std::ios::app | std::ios::binary);
    for (auto &clip : sort_buffer_) {
      for (const auto &col : executor->cols()) {
        ofs.write(clip->data + col.offset, col.len);
      }
    }
    ofs.close();
  }

  bool need_sort(AbstractExecutor *executor, const std::string &idx_name) {
    if (dynamic_cast<SeqScanExecutor *>(executor)) {
      return true;
    }
    auto ix_exe = dynamic_cast<IndexScanExecutor *>(executor);
    if (!ix_exe->user_expected_index(idx_name)) {
      return true;
    }
    return false;
  }

  void print_sorted_results() {
    std::ifstream l_ifs(left_sorted_filename_, std::ios::binary);
    std::ifstream r_ifs(right_sorted_filename_, std::ios::binary);

    std::ofstream ofs("sorted_results.txt");

    print_sorted_results_for_one_side(left_.get(), l_ifs, ofs);
    print_sorted_results_for_one_side(right_.get(), r_ifs, ofs);

    l_ifs.close();
    r_ifs.close();
    ofs.close();
  }

  void print_sorted_results_for_one_side(AbstractExecutor *executor, std::ifstream &ifs, std::ofstream &ofs) {
    std::vector<std::string> captions;
    captions.reserve(executor->cols().size());
    for (auto &sel_col : executor->cols()) {
      captions.push_back(sel_col.name);
    }
    ofs << "|";
    for (const auto &caption : captions) {
      ofs << " " << caption << " |";
    }
    ofs << "\n";

    while (!ifs.eof()) {
      auto record = std::make_unique<RmRecord>(executor->tupleLen());
      for (const auto &col : executor->cols()) {
        ifs.read(record->data + col.offset, col.len);
      }
      if (ifs.eof()) {
        break;
      }

      std::vector<std::string> columns;
      for (auto &col : executor->cols()) {
        std::string col_str;
        char *rec_buf = record->data + col.offset;
        if (col.type == TYPE_INT) {
          col_str = std::to_string(*(int *)rec_buf);
        } else if (col.type == TYPE_FLOAT) {
          col_str = std::to_string(*(float *)rec_buf);
        } else if (col.type == TYPE_STRING) {
          col_str = std::string((char *)rec_buf, col.len);
          col_str.resize(strlen(col_str.c_str()));
        }
        columns.push_back(col_str);
      }
      // print record into file
      ofs << "|";
      for (const auto &column : columns) {
        ofs << " " << column << " |";
      }
      ofs << "\n";
    }
  }
};
