/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#pragma once
#include "execution_defs.h"
#include "execution_manager.h"
#include "executor_abstract.h"
#include "index/ix.h"
#include "system/sm.h"

class UpdateExecutor : public AbstractExecutor {
private:
  TabMeta tab_;
  std::vector<Condition> conds_;
  RmFileHandle* fh_;
  std::vector<Rid> rids_;
  std::string tab_name_;
  std::vector<SetClause> set_clauses_;
  SmManager* sm_manager_;

public:
  UpdateExecutor(SmManager* sm_manager, const std::string& tab_name, std::vector<SetClause> set_clauses,
                 std::vector<Condition> conds, std::vector<Rid> rids, Context* context) {
    sm_manager_ = sm_manager;
    tab_name_ = tab_name;
    set_clauses_ = set_clauses;
    tab_ = sm_manager_->db_.get_table(tab_name);
    fh_ = sm_manager_->fhs_.at(tab_name).get();
    conds_ = conds;
    rids_ = rids;
    context_ = context;
  }

  std::unique_ptr<RmRecord> Next() override {
    //对表上写锁
    if (!context_->lock_mgr_->lock_exclusive_on_table(context_->txn_, fh_->GetFd())) {
      throw TransactionAbortException(context_->txn_->get_transaction_id(), AbortReason::UPGRADE_CONFLICT);
    }

    //更新主表记录
    std::map<std::string, int> mp;
    for (int i = 0; i < tab_.cols.size(); i++) {
      mp[tab_.cols[i].name] = i + 1;
    }
    //标记要更新的index
    std::vector<int> idxUpdate;
    for (int i = 0; i < tab_.indexes.size(); i++) {
      for (auto& set_clause : set_clauses_) {
        bool flag = false;
        for (int j = 0; j < tab_.indexes[i].cols.size(); j++) {
          if (tab_.indexes[i].cols[j].name == set_clause.lhs.col_name) {
            idxUpdate.push_back(i);
            flag = true;
            break;
          }
        }
        if (flag) break;
      }
    }

    for (auto& rid : rids_) {
      auto rec = fh_->get_record(rid, context_);
      auto arc_rec = std::make_unique<RmRecord>(*rec);
      auto pre_rec = new char[fh_->get_tab_record_size()];
      memcpy(pre_rec, rec->data, fh_->get_tab_record_size());//更新之前的数据

      for (auto& set_clause : set_clauses_) {
        //找到字段
        int col_id = mp[set_clause.lhs.col_name] - 1;
        if (col_id == -1) {
          throw ColumnNotFoundError(set_clause.lhs.col_name);
        }
        auto& col = tab_.cols[col_id];
        auto& val = set_clause.rhs;
        //如果rhs是int,lhs为float，则转换
        if (val.type == TYPE_INT && col.type == TYPE_FLOAT) {
          val.set_float((float)val.int_val);
        }
        if (col.type != val.type) {
          if (col.type == TYPE_FLOAT && val.type == TYPE_INT) {
            val.set_float(static_cast<float>(val.int_val));
          } else {
            throw IncompatibleTypeError(coltype2str(col.type), coltype2str(val.type));
          }
        }
        val.init_raw(col.len);
        memcpy(rec->data + col.offset, val.raw->data, col.len);
      }

      for (auto index_meta : tab_.indexes) {
        auto ih = sm_manager_->ihs_.at(sm_manager_->get_ix_manager()->get_index_name(tab_name_, index_meta.cols)).get();

        char* pre_key = new char[index_meta.col_tot_len];
        int offset = 0;
        for (size_t i = 0; i < index_meta.col_num; i++) {
          memcpy(pre_key + offset, pre_rec + index_meta.cols[i].offset, index_meta.cols[i].len);
          offset += index_meta.cols[i].len;
        }

        char* new_key = new char[index_meta.col_tot_len];
        offset = 0;
        for (size_t i = 0; i < index_meta.col_num; i++) {
          memcpy(new_key + offset, rec->data + index_meta.cols[i].offset, index_meta.cols[i].len);
          offset += index_meta.cols[i].len;
        }

        auto cmp = memcmp(pre_key, new_key, index_meta.col_tot_len);
        if (cmp == 0) {
          continue;
        }

        std::vector<Rid> result;
        if (ih->get_value(new_key, &result, context_->txn_)) {
          throw InternalError("Duplicate key for unique index when update!");
        }
      }

      fh_->update_record(rid, rec->data, context_);

      auto write_record = new WriteRecord(WType::UPDATE_TUPLE, tab_name_, rid, *arc_rec);
      context_->txn_->append_record_write_record(write_record);

      //更新索引
      for (auto& idx : idxUpdate) {
        auto& index = tab_.indexes[idx];
        auto ih = sm_manager_->ihs_.at(sm_manager_->get_ix_manager()->get_index_name(tab_name_, index.cols)).get();
        char* old_key = new char[index.col_tot_len];
        char* new_key = new char[index.col_tot_len];

        int offset = 0;
        for (size_t i = 0; i < index.col_num; i++) {
          memcpy(old_key + offset, pre_rec + index.cols[i].offset, index.cols[i].len);
          memcpy(new_key + offset, rec->data + index.cols[i].offset, index.cols[i].len);
          offset += index.cols[i].len;
        }
        ih->delete_entry(old_key, context_->txn_);
        ih->insert_entry(new_key, rid, context_->txn_);

        auto delete_index_write_record = new IndexWriteRecord(WType::DELETE_TUPLE, tab_name_, rid, index.col_tot_len,
                                                              old_key);
        auto insert_index_write_record = new IndexWriteRecord(WType::INSERT_TUPLE, tab_name_, rid, index.col_tot_len,
                                                              new_key);
        context_->txn_->append_index_write_record(delete_index_write_record);
        context_->txn_->append_index_write_record(insert_index_write_record);

        delete[] old_key;
        delete[] new_key;
      }

      delete[] pre_rec;
    }

    mp.clear();

    return nullptr;
  }

  Rid& rid() override { return _abstract_rid; }
};
