/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "lock_manager.h"

#include <common/common.h>

/**
 * @description: 申请行级共享锁
 * @return {bool} 加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {Rid&} rid 加锁的目标记录ID 记录所在的表的fd
 * @param {int} tab_fd
 */
bool LockManager::lock_shared_on_record(Transaction *txn, const Rid &rid, int tab_fd) {
  return true;
}

/**
 * @description: 申请行级排他锁
 * @return {bool} 加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {Rid&} rid 加锁的目标记录ID
 * @param {int} tab_fd 记录所在的表的fd
 */
bool LockManager::lock_exclusive_on_record(Transaction *txn, const Rid &rid, int tab_fd) {
  return true;
}

/**
 * @description: 申请表级读锁
 * @return {bool} 返回加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {int} tab_fd 目标表的fd
 */
bool LockManager::lock_shared_on_table(Transaction *txn, int tab_fd) {
  std::unique_lock<std::mutex> lock(latch_);

  LOG_DEBUG("transaction " + std::to_string(txn->get_transaction_id()) + " requests shared lock on table " +
      std::to_string(tab_fd) + ".");

  // 更新/判断事物状态
  if (txn->get_state() == TransactionState::DEFAULT) {
    txn->set_state(TransactionState::GROWING);
  }
  if (txn->get_state() != TransactionState::GROWING) {
    throw TransactionAbortException(txn->get_transaction_id(), AbortReason::LOCK_ON_SHIRINKING);
  }

  // 组装锁ID和请求
  LockDataId lock_data_id = {tab_fd, LockDataType::TABLE};
  LockRequest request(txn->get_transaction_id(), LockMode::SHARED);
  auto &request_queue = lock_table_[lock_data_id];

  //检查是否持有相同/更高级的锁
  if (has_ge_level_lock_on_same_txn(txn, LockMode::SHARED, request_queue)) {
    return true;
  }

  auto &insert_it = request_queue.request_queue_.emplace_back(request);
  //将锁加入到事务的锁集合中
  txn->get_lock_set()->insert(lock_data_id);

  // 死锁预防
  wait_die(request_queue, insert_it, lock);

  //将请求加入到请求队列中
  insert_it.granted_ = true;
  request_queue.group_lock_mode_ = GroupLockMode::S;

  LOG_DEBUG("transaction " + std::to_string(txn->get_transaction_id()) + " gets shared lock on table " +
      std::to_string(tab_fd) + ".");

  return true;
}

/**
 * @description: 申请表级写锁
 * @return {bool} 返回加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {int} tab_fd 目标表的fd
 */
bool LockManager::lock_exclusive_on_table(Transaction *txn, int tab_fd) {
  std::unique_lock<std::mutex> lock(latch_);

  LOG_DEBUG("transaction " + std::to_string(txn->get_transaction_id()) + " requests exclusive lock on table " +
      std::to_string(tab_fd) + ".");

  // 更新/判断事物状态
  if (txn->get_state() == TransactionState::DEFAULT) {
    txn->set_state(TransactionState::GROWING);
  }
  if (txn->get_state() != TransactionState::GROWING) {
    throw TransactionAbortException(txn->get_transaction_id(), AbortReason::LOCK_ON_SHIRINKING);
  }

  // 组装事物ID和请求
  LockDataId lock_data_id = {tab_fd, LockDataType::TABLE};
  LockRequest request(txn->get_transaction_id(), LockMode::EXLUCSIVE);
  auto &request_queue = lock_table_[lock_data_id];

  //检查是否持有相同/更高级的锁
  if (has_ge_level_lock_on_same_txn(txn, LockMode::EXLUCSIVE, request_queue)) {
    return true;
  }

  //如果有更低级的锁，需要考虑锁升级
  if (has_lt_level_lock_on_same_txn(txn, LockMode::EXLUCSIVE, request_queue)) {
    upgrade_lock_on_table(txn, tab_fd, lock);
    return true;
  }

  // 添加到queue中，此时未grant
  auto &insert_it = request_queue.request_queue_.emplace_back(request);
  txn->get_lock_set()->insert(lock_data_id);

  // 利用wait-die策略确保不会产生死锁，以及事物可以获得锁
  wait_die(request_queue, insert_it, lock);

  insert_it.granted_ = true;
  request_queue.group_lock_mode_ = GroupLockMode::X;

  LOG_DEBUG("transaction " + std::to_string(txn->get_transaction_id()) + " gets exclusive lock on table " +
      std::to_string(tab_fd) + ".");

  return true;
}

/**
 * @description: 申请表级意向读锁
 * @return {bool} 返回加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {int} tab_fd 目标表的fd
 */
bool LockManager::lock_IS_on_table(Transaction *txn, int tab_fd) {
  return true;
}

/**
 * @description: 申请表级意向写锁
 * @return {bool} 返回加锁是否成功
 * @param {Transaction*} txn 要申请锁的事务对象指针
 * @param {int} tab_fd 目标表的fd
 */
bool LockManager::lock_IX_on_table(Transaction *txn, int tab_fd) {
  return true;
}

/**
 * @description: 释放锁
 * @return {bool} 返回解锁是否成功
 * @param {Transaction*} txn 要释放锁的事务对象指针
 * @param {LockDataId} lock_data_id 要释放的锁ID
 */
bool LockManager::unlock(Transaction *txn, LockDataId lock_data_id) {
  std::unique_lock<std::mutex> lock(latch_);

  auto &request_queue = lock_table_[lock_data_id];
  auto &request_queue_ = request_queue.request_queue_;
  auto it = std::find_if(request_queue_.begin(), request_queue_.end(),
                         [txn](const LockRequest &req) {
                           return req.txn_id_ == txn->get_transaction_id();
                         });

  auto mode_str = "";
  if (it->lock_mode_ == LockMode::EXLUCSIVE) {
    mode_str = "X";
  } else {
    mode_str = "S";
  }
  LOG_DEBUG("transaction " + std::to_string(txn->get_transaction_id()) + " releases lock on table " +
      std::to_string(lock_data_id.fd_) + ",mode is " +mode_str + ".");

  request_queue_.erase(it);

  auto lock_set = txn->get_lock_set();
  lock_set->erase(lock_data_id);

  auto s_it = std::find_if(request_queue_.begin(), request_queue_.end(),
                           [txn](const LockRequest &req) {
                             return req.txn_id_ == txn->get_transaction_id() && req.lock_mode_ == LockMode::SHARED &&
                                    req.granted_;
                           });
  auto x_it = std::find_if(request_queue_.begin(), request_queue_.end(),
                           [txn](const LockRequest &req) {
                             return req.txn_id_ == txn->get_transaction_id() && req.lock_mode_ == LockMode::EXLUCSIVE &&
                                    req.granted_;
                           });
  if (s_it == request_queue_.end() && x_it == request_queue_.end()) {
    auto new_x_it = std::find_if(request_queue_.begin(), request_queue_.end(),
                                 [txn](const LockRequest &req) {
                                   return req.txn_id_ != txn->get_transaction_id() && req.lock_mode_ ==
                                          LockMode::EXLUCSIVE && !req.granted_;
                                 });
    if (new_x_it != request_queue_.end()) {
      new_x_it->granted_ = true;
      request_queue.cv_.notify_all();
      request_queue.group_lock_mode_ = GroupLockMode::X;
      return true;
    }
  }
  if (x_it == request_queue_.end()) {
    auto cnt = 0;
    for (auto &new_x_it : request_queue_) {
      if (new_x_it.txn_id_ != txn->get_transaction_id() && new_x_it.lock_mode_ == LockMode::SHARED && !new_x_it.
          granted_) {
        new_x_it.granted_ = true;
        request_queue.cv_.notify_all();
        cnt++;
      }
    }
    if (cnt != 0) {
      request_queue.group_lock_mode_ = GroupLockMode::S;
      return true;
    }
  }

  request_queue.group_lock_mode_ = GroupLockMode::NON_LOCK;
  return true;
}

void LockManager::upgrade_lock_on_table(Transaction *txn, int tab_fd, std::unique_lock<std::mutex> &queue_lock) {
  LockDataId lock_data_id = {tab_fd, LockDataType::TABLE};
  auto &request_queue = lock_table_[lock_data_id];

  auto it = std::find_if(request_queue.request_queue_.begin(), request_queue.request_queue_.end(),
                         [txn](const LockRequest &req) {
                           return req.txn_id_ == txn->get_transaction_id() && req.lock_mode_ == LockMode::SHARED;
                         });
  if (it != request_queue.request_queue_.end()) {
    // 找到共享锁，尝试释放并升级到排他锁
    it->granted_ = false; // 设置共享锁未授予状态，开始释放共享锁
    request_queue.request_queue_.erase(it); // 从队列中移除共享锁请求

    // 创建排他锁请求
    LockRequest new_request(txn->get_transaction_id(), LockMode::EXLUCSIVE);
    auto &insert_it = request_queue.request_queue_.emplace_back(new_request);

    // 使用 wait-die 策略等待排他锁
    wait_die(request_queue, insert_it, queue_lock);

    // 排他锁授予成功
    request_queue.group_lock_mode_ = GroupLockMode::X;
    insert_it.granted_ = true;

    return;
  }
  throw InternalError("No shared lock found for upgrade at txn: " + std::to_string(txn->get_transaction_id()));
}

bool LockManager::has_lt_level_lock_on_same_txn(Transaction *txn, LockMode target_mode,
                                                LockRequestQueue &request_queue) {
  for (auto &req : request_queue.request_queue_) {
    if (req.txn_id_ != txn->get_transaction_id()) {
      continue;
    }
    //先直接判断，不考虑共享锁
    if (target_mode == LockMode::EXLUCSIVE && req.lock_mode_ == LockMode::SHARED) {
      return true;
    }
  }
  return false;
}

bool LockManager::has_ge_level_lock_on_same_txn(Transaction *txn, LockMode target_mode,
                                                LockRequestQueue &request_queue) {
  //只考虑表锁
  for (auto &req : request_queue.request_queue_) {
    if (req.txn_id_ != txn->get_transaction_id()) {
      continue;
    }
    if (target_mode == LockMode::SHARED &&
        (req.lock_mode_ == LockMode::SHARED || req.lock_mode_ == LockMode::EXLUCSIVE)) {
      return true;
    }
    if (target_mode == LockMode::EXLUCSIVE && req.lock_mode_ == LockMode::EXLUCSIVE) {
      return true;
    }
  }
  return false;
}

/*
 * @description: 实现wait-die，调用此方法后当前事物已经可以获得锁
 * todo：释放s锁-获得x锁可能会有并发问题
 * @param {LockRequestQueue} request_queue
 * @param {LockRequest} request
 * @param {std::unique_lock<std::mutex>} lock
 */
void LockManager::wait_die(LockRequestQueue &request_queue, LockRequest &request, std::unique_lock<std::mutex> &lock) {
  // 用于记录冲突数
  int conflict_count = 0;
  for (const auto &req : request_queue.request_queue_) {
    if (req.txn_id_ == request.txn_id_) {
      // 同一事务的其他请求跳过
      continue;
    }

    // 检查锁兼容性
    if (!are_locks_compatible(req.lock_mode_, request.lock_mode_)) {
      if (req.txn_id_ < request.txn_id_) {
        // 请求的事物如果是较年轻事物，则回滚
        throw TransactionAbortException(request.txn_id_, AbortReason::DEADLOCK_PREVENTION);
      }
      conflict_count++;
    }
  }

  // 如果存在冲突，则等待
  if (conflict_count > 0) {
    request_queue.cv_.wait(lock, [&request] {
      LOG_DEBUG("transaction " + std::to_string(request.txn_id_) + " check for lock and grant_ is "
          + (request.granted_ ?"true":"false") + " .");
      return request.granted_;
    });
  }
}

/*
 * @description: 判断当前的等待队列是否可以让指定的request获得锁
 */
bool LockManager::are_locks_compatible(LockMode existing, LockMode requested) {
  // 兼容性逻辑实现
  if (existing == LockMode::SHARED && requested == LockMode::SHARED) {
    return true; // 共享锁之间是兼容的
  }
  return false; // 其他情况均不兼容
}