/* Copyright (c) 2023 Renmin University of China
RMDB is licensed under Mulan PSL v2.
You can use this software according to the terms and conditions of the Mulan PSL v2.
You may obtain a copy of Mulan PSL v2 at:
        http://license.coscl.org.cn/MulanPSL2
THIS SOFTWARE IS PROVIDED ON AN "AS IS" BASIS, WITHOUT WARRANTIES OF ANY KIND,
EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO NON-INFRINGEMENT,
MERCHANTABILITY OR FIT FOR A PARTICULAR PURPOSE.
See the Mulan PSL v2 for more details. */

#include "transaction_manager.h"

#include <common/common.h>

#include "record/rm_file_handle.h"
#include "system/sm_manager.h"

std::unordered_map<txn_id_t, Transaction *> TransactionManager::txn_map = {};

/**
 * @description: 事务的开始方法
 * @return {Transaction*} 开始事务的指针
 * @param {Transaction*} txn 事务指针，空指针代表需要创建新事务，否则开始已有事务
 * @param {LogManager*} log_manager 日志管理器指针
 */
Transaction *TransactionManager::begin(Transaction *txn, LogManager *log_manager) {
  // 1. 判断传入事务参数是否为空指针
  // 2. 如果为空指针，创建新事务
  // 3. 把开始事务加入到全局事务表中
  // 4. 返回当前事务指针
  if (txn == nullptr) {
    txn = new Transaction(next_txn_id_, IsolationLevel::SERIALIZABLE);
    next_txn_id_++;
    txn->set_start_ts(next_timestamp_);
    next_timestamp_++;
  }
  std::unique_lock<std::mutex> lock(latch_);
  txn_map[txn->get_transaction_id()] = txn;
  txn->set_state(TransactionState::DEFAULT);
  lock.unlock();

  LOG_INFO("transaction " + std::to_string(txn->get_transaction_id()) + " begin.");

  return txn;
}

/**
 * @description: 事务的提交方法
 * @param {Transaction*} txn 需要提交的事务
 * @param {LogManager*} log_manager 日志管理器指针
 */
void TransactionManager::commit(Transaction *txn, LogManager *log_manager) {
  // 1. 如果存在未提交的写操作，提交所有的写操作
  // 2. 释放所有锁
  // 3. 释放事务相关资源，eg.锁集 todo
  // 4. 把事务日志刷入磁盘中
  // 5. 更新事务状态

  std::scoped_lock lock(latch_);

  release_all_locks(txn);

  auto record_write_set = txn->get_record_write_set();
  record_write_set->clear();
  auto index_write_set = txn->get_index_write_set();
  index_write_set->clear();

  txn->set_state(TransactionState::COMMITTED);

  LOG_INFO("transaction " + std::to_string(txn->get_transaction_id()) + " commit.");
}

/**
 * @description: 事务的终止（回滚）方法
 * @param {Transaction *} txn 需要回滚的事务
 * @param {LogManager} *log_manager 日志管理器指针
 */
void TransactionManager::abort(Transaction *txn, LogManager *log_manager) {
  // 1. 回滚所有写操作
  // 2. 释放所有锁
  // 3. 清空事务相关资源，eg.锁集
  // 4. 把事务日志刷入磁盘中
  // 5. 更新事务状态

  std::scoped_lock lock(latch_);
  //todo：暂时设置为null
  Context *context = nullptr;

  release_all_locks(txn);

  auto record_write_set = txn->get_record_write_set();
  while (!record_write_set->empty()) {
    auto record = record_write_set->back();
    auto w_type = record->GetWriteType();
    auto tab_name = record->GetTableName();
    auto rm_file = sm_manager_->fhs_.at(tab_name).get();
    if (w_type == WType::INSERT_TUPLE) {
      rm_file->delete_record(record->GetRid(), context);
    } else if (w_type == WType::UPDATE_TUPLE) {
      rm_file->update_record(record->GetRid(), record->GetRecord().data, context);
    } else {
      rm_file->insert_record(record->GetRecord().data, context);
    }
    record_write_set->pop_back();
    delete record;
  }
  record_write_set->clear();

  auto index_write_set = txn->get_index_write_set();
  while (!index_write_set->empty()) {
    auto record = index_write_set->back();
    auto w_type = record->GetWriteType();
    auto indexes = sm_manager_->db_.get_table(record->GetTableName()).indexes;
    if (w_type == WType::INSERT_TUPLE) {
      for (const auto &index_meta : indexes) {
        auto idx_name = sm_manager_->get_ix_manager()->get_index_name(record->GetTableName(), index_meta.cols);
        auto ih = sm_manager_->ihs_.at(idx_name).get();
        ih->delete_entry(record->GetKey(), txn);
      }
    } else {
      for (auto &index_meta : indexes) {
        auto idx_name = sm_manager_->get_ix_manager()->get_index_name(record->GetTableName(), index_meta.cols);
        auto ih = sm_manager_->ihs_.at(idx_name).get();
        ih->insert_entry(record->GetKey(), record->GetRid(), txn);
      }
    }
    index_write_set->pop_back();
    delete record;
  }
  index_write_set->clear();

  txn->set_state(TransactionState::ABORTED);

  LOG_INFO("transaction " + std::to_string(txn->get_transaction_id()) + " abort.");
}

void TransactionManager::release_all_locks(Transaction *txn) {
  std::vector<LockDataId> lock_ids;
  std::copy(txn->get_lock_set()->begin(), txn->get_lock_set()->end(), std::back_inserter(lock_ids));
  for (auto &lock_id : lock_ids) {
    lock_manager_->unlock(txn, lock_id);
  }
}